#include "LocBra.h"

int main(int argc, char **argv)
{                                       /* Local Branching, code after: - the paper writing 
                                                                        - callbackfunctions */
                                        /* Andrea, Matteo: May 2002 */

   int        i, nIter, maxIter, status, Improved, lpstat, Optimum, div, heurFreq;
   int        first, nSolLim, type, presolve, exact, refine, video, nIntervals;
   double     Time, timeLimit, time4Neig, t4N, t_end; 
   char       fname[40];
   double     xBarValue, bestValue, bestBound, rhs;
   int        emphasis, nNodes, prec, mipInt;
 
   double    *xBar;
   char      *ctype;

   if (argc != 14) {
      printf("\n Usage: %s totalTime time4N emp pre heurFreq prec", argv[0]); 
      printf(" mipInt kSize original exact istName");
      printf("\n      where:     totalTime  -> given TimeLimit");
      printf("\n                    time4N  -> time for Neighborhood");
      printf("\n                       emp  -> emphasis: OPT = 0, FEAS = 1");
      printf("\n                       pre  -> presolve: 1 yes, 0 no");
      printf("\n                  heurFreq  -> heuristic frequency");
      printf("\n                      prec  -> precision: 0 = default; 1 = 1e-12");
      printf("\n                    mipInt  -> print MIP interval");
      printf("\n                   timeInt  -> number of time intervals");
      printf("\n                         k  -> size");
      printf("\n                  original  -> original: 1 old version, 0 new version");
      printf("\n                     exact  -> exact: 1 exact, 0 heuristic");
      printf("\n                     video  -> video: YES = 1, NO = 0");
      printf("\n                   istName  -> instance name");
      printf("\n");
      exit (0);
   }
   sscanf(argv[1], "%lf", &timeLimit);
   sscanf(argv[2], "%lf", &time4Neig);
   sscanf(argv[3], "%d",  &emphasis);
   sscanf(argv[4], "%d",  &presolve);
   sscanf(argv[5], "%d",  &heurFreq);
   sscanf(argv[6], "%d",  &prec);
   sscanf(argv[7], "%d",  &mipInt);
   sscanf(argv[8], "%d",  &nIntervals);
   sscanf(argv[9], "%d",  &kSize);
   sscanf(argv[10], "%d", &original);
   sscanf(argv[11], "%d", &exact);
   sscanf(argv[12], "%d", &video);
   sscanf(argv[13], "%s", fname);

   printf("\n LOCBRA timeLimit, time4Neig, emphasis, presolve, heurFreq");
   printf("\n        prec, mipInt, timeInt, kSize, original, exact, video, fname");
   printf("\n LOCBRA %10.0f %10.0f %d %d %d %d %d %d %d %d %d %d %s \n", 
    timeLimit, time4Neig, emphasis, presolve, heurFreq, prec, mipInt, nIntervals, kSize, original, exact, video, fname);

/* initialize environment variables */
   env   = NULL;
   lp    = NULL;
   Eps   = .00001;
   Inf   = 1000000000; 
   dInf  = 100000000000.; 

/* optimization parameters */
   if (exact) dmax = 0;
   else       dmax = Inf;
   maxIter = Inf;
   div     = 0;
   first   = 1;
   refine  = 0;
   debug   = 0;
   if (timeLimit < 0.) timeLimit = dInf;
   timeSlot = timeLimit / nIntervals;
   slot = (int*)calloc(nIntervals, sizeof(int)); slot[0] = 1;

/* CPLEX initialization */
   env = CPXopenCPLEXdevelop(&status);
   if (env == NULL) printf("\n error CPXopenCPLEXdevelop");
   lp = CPXcreateprob(env, &status, "LocBra");
   if (lp == NULL) printf("\n error CPXcreateprob");

/* read the problem */
   status = CPXreadcopyprob(env, lp, fname, "LP");
   if (status) printf("\n error CPXreadcopyprob, status = %d", status);

/* get information about the problem */
   MinMax  = CPXgetobjsen(env, lp);
   ncols = CPXgetnumcols(env, lp);
   nrows = CPXgetnumrows(env, lp);

/* memory allocation */
   xBar = (double*)calloc(ncols, sizeof(double));
   ctype = (char*)calloc(ncols, sizeof(char));

   status = CPXgetctype(env, lp, ctype, 0, ncols - 1);
   nBcols = 0; for (i = 0; i < ncols; i++) if (ctype[i] == 'B') nBcols++;

#ifdef VERBOSE
   printf("\n Problem info: MinMax = %d, ncols = %d, nBcols = %d, nrows = %d\n", MinMax, ncols, nBcols, nrows);
#endif

/* initialization */
   nIter = dv = 0; Optimum = 0;
   t_start = second(); Time = 0.;
   overallBestValue = bestValue = (double)MinMax * dInf;
   t4N = dInf;

/* cplex paramenters */
   if (prec) {
      status = CPXsetdblparam(env, CPX_PARAM_EPAGAP, 1e-12);
      status = CPXsetdblparam(env, CPX_PARAM_EPGAP, 1e-06);
   }
   status = CPXsetintparam(env, CPX_PARAM_SCRIND, video);
   status = CPXsetintparam(env, CPX_PARAM_MIPINTERVAL, mipInt);
   status = CPXsetintparam(env, CPX_PARAM_HEURFREQ, heurFreq);
   status = CPXsetintparam(env, CPX_PARAM_MIPEMPHASIS, emphasis);
   status = CPXsetintparam(env, CPX_PARAM_PREIND, presolve);

   status = CPXsetmipcallbackfunc(env, mycallback, NULL);
   if (status) {
      printf ("\nError in CPXsetlpcallbackfunc: %d", status);
      exit(1);
   }

   while (Time < timeLimit && nIter < maxIter && !Optimum) {
         printf("\n CPLEX call # %d started!\n", nIter + 1);

/* set TIME limit for each neighborhood exploration */
         if (t4N > timeLimit - Time) t4N = timeLimit - Time;
         status = CPXsetdblparam(env, CPX_PARAM_TILIM, t4N);

/* set first/best neighborhood improvement option */
         nSolLim = Inf;
         if (first) nSolLim = 1;
         status = CPXsetintparam(env, CPX_PARAM_INTSOLLIM, nSolLim);

/* mipopt */
         status = CPXmipopt(env, lp);
         if (status) printf("\nError in CPXmipopt: %d", status);
         t4N = time4Neig;

/* check lpstat */
         if ((lpstat = CPXgetstat(env, lp))) printf("\n   Optimization status = %d", lpstat);
         else { printf("\nSomething wrong: lpstat = %d", lpstat); exit(0); }

/* 1 */  if (lpstat == 101 || lpstat == 102) {
            xBarValue = getSolution(xBar);
            Improved = checkSol(xBarValue, &bestValue);
            if (!Improved) {
               printf("\n   an improvement was expected: something wrong!");
               exit(1);
            }
            printf("\n   IMPROVED solution = %10.2f", bestValue);
            if (!nIter) { Optimum = 1; goto end; }
            status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
            div = first = 0; rhs = kSize; 
            type = 0; modifyRow(type);
          }    
/* 2 */  if (lpstat == 103) {
            if (!nIter) { 
               Optimum = 1; 
               printf("\n   NO FEASIBLE solution exists");
               goto end; 
            }
            type = 0; modifyRow(type);   
            if (div || exact) { bestValue = t4N = dInf; dv++; first = 1; }
            status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
            div = 1; rhs += kSize/2;
         }
/* 3 */  if (lpstat == 104 || lpstat == 107) {
            xBarValue = getSolution(xBar);
            if (nIter) {
               if (lpstat == 104) {                  // last_mod
                  type = 2; modifyRow(type); 
               } else {
                  if (lpstat == 107) {               // last_mod
                     type = 1; modifyRow(type); 
                  } else { 
                     printf("\n strange lpstat"); exit(0); 
                  }
               }
            } else {
               status = CPXgetbestobjval(env, lp, &bestBound);
               if (status) printf("\nError in CPXgetbestobjval: %d", status);
               nNodes = CPXgetnodecnt(env, lp);
               printf("\n   Final bound = %10.2f after %d B&B nodes", bestBound, nNodes);
            }
            if (ncols != nBcols) {
               printf("\n   IMPROVED solution before REFINEMENT = %10.2f\n", xBarValue);
               refineSolution(&xBarValue, xBar);
            }
            Improved = checkSol(xBarValue, &bestValue);
            if (!Improved) {
               printf("\n   an improvement was expected: something wrong!");
               exit(1);
            }
            printf("\n   IMPROVED solution = %10.2f", bestValue);
            status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
            div = first = 0; rhs = kSize;
         }
/* 4 */  if (lpstat == 108) {
            if (nIter && !div) { type = 2; modifyRow(type); } // last_mod
            else               { type = 1; modifyRow(type);   // last_mod
                                 if (!nIter) { printf("\n time limit and no sol!"); exit(0); } }
               
            if (div || exact) { bestValue = t4N = dInf; dv++; first = 1; rhs += kSize/2; } 
            else rhs -= kSize/2;
            status = CPXsetdblparam(env, CPX_PARAM_CUTUP, bestValue - Eps);
            div = 1;
         }
        
         if (dv > dmax) { exact = 1; goto out; }

         if (debug == 1) status = CPXwriteprob(env, lp, "prima.lp", "LP");
         localBranch(ctype, xBar, refine, rhs);
         if (debug == 1) status = CPXwriteprob(env, lp, "dopo.lp", "LP");

end:
         t_end = second(); Time = t_end - t_start;
         printf("\n c. %10.2f - best %10.2f - elapsed time %8.1f (%d)\n", 
                 bestValue, overallBestValue, Time, dv);
         nIter++;
   }

out:
   if (exact && !Optimum) {
      t_end = second(); Time = t_end - t_start;
      printf("\n   Final locBra sol %10.2f - elapsed time %8.1f \n", overallBestValue, Time);
      status = CPXsetdblparam(env, CPX_PARAM_CUTUP, overallBestValue - Eps);
      status = CPXsetdblparam(env, CPX_PARAM_TILIM, timeLimit - Time);
      status = CPXsetintparam(env, CPX_PARAM_INTSOLLIM, Inf);
      status = CPXmipopt(env, lp);
      t_end = second(); Time = t_end - t_start;
      
      if ((lpstat = CPXgetstat(env, lp))) printf("\n   Exact solution status = %d", lpstat);
      else { printf("\nSomething wrong in exact solution: lpstat = %d", lpstat); exit(0); }

      printf("\n   EXACT step performed:");
      switch(lpstat) {
      case 101:
                xBarValue = getSolution(xBar);
                printf("  OPTIMAL sol found %10.2f - elapsed time %8.1f", xBarValue, Time);
                break;
      case 102:
                xBarValue = getSolution(xBar);
                printf("  OPTIMAL sol found %10.2f - elapsed time %8.1f", xBarValue, Time);
                break;
      case 103:
                printf("  OPTIMAL sol proved %10.2f - elapsed time %8.1f", overallBestValue, Time);
                break;
      case 107:
                xBarValue = getSolution(xBar);
                printf("  IMPROVED sol %10.2f - elapsed time %8.1f", xBarValue, timeLimit);
                break;
      case 108:
                printf("  NO IMPROVED sol %10.2f - elapsed time %8.1f", overallBestValue, timeLimit);
                break;
      }
   } else {
      if (!Optimum) printf("\n Final bound = %10.2f after %d B&B nodes", bestBound, nNodes);
      printf("\n Final Solution = %10.2f after %8.1f sec and %d(/%d) div", 
                 overallBestValue, time4Best, dBest, dv);

      if (Time > timeLimit) printf("\n Time Limit of %8.1f reached", timeLimit);
      else if (nIter == maxIter) printf("\n Max Iteration limit of %d reached", maxIter);
   }

   free(xBar); free(ctype); free(slot);
   printf("\n");

   CPXcloseCPLEX(&env);
   return 0;
}
